"""
.. module: historical_reports.s3.generate
    :platform: Unix
    :copyright: (c) 2017 by Netflix Inc., see AUTHORS for more
    :license: Apache, see LICENSE for more details.
.. author:: Mike Grima <mgrima@netflix.com>
"""
import json
import logging

from historical.s3.models import CurrentS3Model

from historical_reports.s3.models import S3ReportSchema
from historical_reports.s3.util import dump_to_s3

logging.basicConfig()
log = logging.getLogger('historical-reports-s3')
log.setLevel(logging.WARNING)


def dump_report(commit=True):
    # Get all the data from DynamoDB:
    log.debug("Starting... Beginning scan.")
    all_buckets = CurrentS3Model.scan()

    generated_file = S3ReportSchema(strict=True).dump({"all_buckets": all_buckets}).data

    # Dump to S3:
    if commit:
        log.debug("Saving to S3.")

        # Replace <empty> with "" <-- Due to Pynamo/Dynamo issues...
        dump_to_s3(json.dumps(generated_file, indent=4).replace("\"<empty>\"", "\"\"").encode("utf-8"))
    else:
        log.debug("Commit flag not set, not saving.")

    log.debug("Completed S3 report generation.")



"""
.. module: historical_reports.s3.tests.conftest
    :platform: Unix
    :copyright: (c) 2017 by Netflix Inc., see AUTHORS for more
    :license: Apache, see LICENSE for more details.
.. author:: Mike Grima <mgrima@netflix.com>
"""
import json

import boto3
import pytest
from historical.s3.models import CurrentS3Model
from historical_reports.s3.models import S3ReportSchema
from moto import mock_dynamodb2, mock_s3
from historical_reports.s3.tests.factories import DynamoDBRecordsFactory, DynamoDBRecordFactory, DynamoDBDataFactory, serialize, \
    UserIdentityFactory

S3_BUCKET = """{
    "arn": "arn:aws:s3:::testbucket{number}",
    "principalId": "joe@example.com",
    "userIdentity": {
        "sessionContext": {
            "userName": "TseXSKEYQrxm",
            "type": "Role",
            "arn": "arn:aws:iam::123456789012:role/historical_poller",
            "principalId": "AROAIKELBS2RNWG7KASDF",
            "accountId": "123456789012"
        },
        "principalId": "AROAIKELBS2RNWG7KASDF:joe@example.com",
        "type": "Service"
    },
    "accountId": "123456789012",
    "eventTime": "2017-11-10T18:33:44Z",
    "BucketName": "testbucket{number}",
    "Region": "us-east-1",
    "Tags": {
        "theBucketName": "testbucket{number}"
    },
    "configuration": {
        "Grants": {
            "75aa57f09aa0c8caeab4f8c24e99d10f8e7faeebf76c078efc7c6caea54ba06a": [
                "FULL_CONTROL"
            ]
        },
        "LifecycleRules": [
            {
                "Expiration": {
                    "Date": "2015-01-01T00:00:00Z",
                    "Days": 123
                },
                "ID": "string",
                "Prefix": "string",
                "Status": "Enabled",
                "Transitions": [
                    {
                        "Date": "2015-01-01T00:00:00Z",
                        "Days": 123,
                        "StorageClass": "GLACIER"
                    }
                ]
            }
        ],
        "Logging": {},
        "Policy": null,
        "Tags": {
            "theBucketName": "testbucket{number}"
        },
        "Versioning": {},
        "Website": null,
        "Cors": [],
        "Notifications": {},
        "Acceleration": null,
        "Replication": {},
        "CreationDate": "2006-02-03T16:45:09Z",
        "AnalyticsConfigurations": [],
        "MetricsConfigurations": [],
        "InventoryConfigurations": [],
        "Name": "testbucket{number}",
        "_version": 8
    }
}"""


@pytest.fixture(scope='function')
def dynamodb():
    with mock_dynamodb2():
        yield boto3.client('dynamodb', region_name='us-east-1')


@pytest.fixture(scope='function')
def s3():
    with mock_s3():
        yield boto3.client('s3', region_name='us-east-1')


@pytest.fixture(scope='function')
def current_s3_table(dynamodb):
    yield CurrentS3Model.create_table(read_capacity_units=1, write_capacity_units=1, wait=True)


@pytest.fixture(scope="function")
def historical_table(current_s3_table):
    for x in range(0, 10):
        bucket = json.loads(S3_BUCKET.replace("{number}", "{}".format(x)))
        CurrentS3Model(**bucket).save()


@pytest.fixture(scope="function")
def dump_buckets(s3):
    for x in range(0, 10):
        s3.create_bucket(Bucket="dump{}".format(x))

    return s3


@pytest.fixture(scope="function")
def generated_file(historical_table):
    all_buckets = CurrentS3Model.scan()
    return S3ReportSchema(strict=True).dumps({"all_buckets": all_buckets}).data.encode("utf-8")


@pytest.fixture(scope="function")
def generated_report(generated_file):
    return json.loads(generated_file.decode("utf-8"))


@pytest.fixture(scope="function")
def bucket_event():
    new_bucket = json.loads(S3_BUCKET.replace("{number}", "NEWBUCKET"))

    new_item = DynamoDBRecordsFactory(
        records=[
            DynamoDBRecordFactory(
                dynamodb=DynamoDBDataFactory(
                    NewImage=new_bucket,
                    Keys={
                        'arn': new_bucket['arn']
                    }
                ),
                eventName='INSERT'
            )
        ]
    )
    return json.loads(json.dumps(new_item, default=serialize))


@pytest.fixture(scope="function")
def delete_bucket_event():
    delete_bucket = json.loads(S3_BUCKET.replace("{number}", "NEWBUCKET"))
    delete_bucket["configuration"] = {}

    new_item = DynamoDBRecordsFactory(
        records=[
            DynamoDBRecordFactory(
                dynamodb=DynamoDBDataFactory(
                    NewImage=delete_bucket,
                    Keys={
                        'arn': delete_bucket['arn']
                    }
                ),
                eventName='MODIFY'
            )
        ]
    )
    return json.loads(json.dumps(new_item, default=serialize))


@pytest.fixture(scope="function")
def ttl_event():
    bucket = json.loads(S3_BUCKET.replace("{number}", "NEWBUCKET"))

    new_item = DynamoDBRecordsFactory(
        records=[
            DynamoDBRecordFactory(
                dynamodb=DynamoDBDataFactory(
                    OldImage=bucket,
                    Keys={
                        'arn': bucket['arn']
                    }
                ),
                eventName='REMOVE',
                userIdentity=UserIdentityFactory(
                    type='Service',
                    principalId='dynamodb.amazonaws.com'
                )
            )
        ]
    )
    return json.loads(json.dumps(new_item, default=serialize))


@pytest.fixture(scope="function")
def existing_s3_report(dump_buckets, generated_file):
    dump_buckets.put_object(Bucket="dump0", Key="historical-s3-report.json", Body=generated_file)

    return generated_file



"""
.. module: historical_reports.s3.update
    :platform: Unix
    :copyright: (c) 2017 by Netflix Inc., see AUTHORS for more
    :license: Apache, see LICENSE for more details.
.. author:: Mike Grima <mgrima@netflix.com>
"""
import json
import logging

from historical.attributes import decimal_default
from historical.common.dynamodb import deserialize_current_dynamo_to_pynamo
from historical.s3.models import CurrentS3Model

from historical_reports.s3.generate import dump_report
from historical_reports.s3.models import S3ReportSchema
from historical_reports.s3.util import fetch_from_s3, dump_to_s3
from historical_reports.s3.config import CONFIG

logging.basicConfig()
log = logging.getLogger('historical-reports-s3')
log.setLevel(logging.INFO)


def process_dynamodb_record(record, s3_report):
    """Processes a group of DynamoDB NewImage records."""
    if record['eventName'] in ['INSERT', 'MODIFY']:
        # Serialize that specific report:
        modifed_bucket = deserialize_current_dynamo_to_pynamo(record, CurrentS3Model)

        # Is this a soft-deletion? (Config set to {})?
        if not modifed_bucket.configuration.attribute_values:
            s3_report["buckets"].pop(modifed_bucket.BucketName, None)
        else:
            s3_report["all_buckets"].append(modifed_bucket)

    if record['eventName'] == 'REMOVE':
        # We are *ONLY* tracking the deletions from the DynamoDB TTL service.
        # Why? Because when we process deletion records, we are first saving a new "empty" revision to the "Current"
        # table. The "empty" revision will then trigger this Lambda as a "MODIFY" event. Then, right after it saves
        # the "empty" revision, it will then delete the item from the "Current" table. At that point,
        # we have already saved the "deletion revision" to the "Historical" table. Thus, no need to process
        # the deletion events -- except for TTL expirations (which should never happen -- but if they do, you need
        # to investigate why...)
        if record.get('userIdentity'):
            if record['userIdentity']['type'] == 'Service':
                if record['userIdentity']['principalId'] == 'dynamodb.amazonaws.com':
                    log.debug("Processing TTL deletion.")
                    s3_report["buckets"].pop(record['dynamodb']['OldImage']["BucketName"]["S"], None)


def update_records(records, commit=True):
    log.debug("Starting Record Update.")

    # First, grab the existing JSON from S3:
    existing_json = fetch_from_s3()
    log.debug("Grabbed all the existing data from S3.")

    # If the existing JSON is not present for some reason, then...
    if not existing_json:
        if commit and CONFIG.export_if_missing:
            CONFIG.dump_to_buckets = CONFIG.import_bucket.split(",")
            CONFIG.dump_to_prefix = CONFIG.import_prefix
            log.info("The report does not exist. Dumping the full report to {}/{}".format(CONFIG.import_bucket,
                                                                                          CONFIG.import_prefix))
            dump_report()

        else:
            log.error("[X] The existing log was not present and the `EXPORT_IF_MISSING` env var was "
                      "not set so exiting.")

        return

    # Deserialize the report:
    report = S3ReportSchema().loads(existing_json).data
    report["all_buckets"] = []

    for record in records:
        process_dynamodb_record(record, report)

    # Serialize the data:
    generated_file = S3ReportSchema(strict=True).dump(report).data

    # Dump to S3:
    if commit:
        log.debug("Saving to S3.")

        # Replace <empty> with "" <-- Due to Pynamo/Dynamo issues...
        dump_to_s3(json.dumps(generated_file, indent=4, default=decimal_default).replace("\"<empty>\"", "\"\"").encode(
            "utf-8"))
    else:
        log.debug("Commit flag not set, not saving.")

    log.debug("Completed S3 report update.")



"""
.. module: historical_reports.s3.config
    :platform: Unix
    :copyright: (c) 2017 by Netflix Inc., see AUTHORS for more
    :license: Apache, see LICENSE for more details.
.. author:: Mike Grima <mgrima@netflix.com>
"""
import os


class Config:
    """
    Class for maintaining the configuration for the entire runtime.
    By default, this will set most of the values from the environment variables -- but is also
    configurable via other means.

    Simply import this and use it:
    ```
    from s3.config import CONFIG
    ```
    """
    def __init__(self):
        self._s3_reports_version = 1    # Read-only -- this will update as the code changes.
        self._current_region = os.environ.get("AWS_DEFAULT_REGION", "us-east-1")
        self._exclude_fields = os.environ.get("EXCLUDE_FIELDS", "Name,_version").split(",")
        self._dump_to_buckets = os.environ.get("DUMP_TO_BUCKETS", "").split(",")
        self._dump_to_prefix = os.environ.get("DUMP_TO_PREFIX", "historical-s3-report.json")
        self._import_bucket = os.environ.get("IMPORT_BUCKET", None)
        self._import_prefix = os.environ.get("IMPORT_PREFIX", "historical-s3-report.json")
        self._export_if_missing = os.environ.get("EXPORT_IF_MISSING", False)

    @property
    def s3_reports_version(self):
        return self._s3_reports_version

    @property
    def current_region(self):
        return self._current_region

    @current_region.setter
    def current_region(self, region):
        self._current_region = region

    @property
    def exclude_fields(self):
        return self._exclude_fields

    @exclude_fields.setter
    def exclude_fields(self, fields):
        self._exclude_fields = fields

    @property
    def dump_to_buckets(self):
        return self._dump_to_buckets

    @dump_to_buckets.setter
    def dump_to_buckets(self, buckets):
        self._dump_to_buckets = buckets

    @property
    def dump_to_prefix(self):
        return self._dump_to_prefix

    @dump_to_prefix.setter
    def dump_to_prefix(self, prefix):
        self._dump_to_prefix = prefix

    @property
    def import_bucket(self):
        return self._import_bucket

    @import_bucket.setter
    def import_bucket(self, bucket):
        self._import_bucket = bucket

    @property
    def import_prefix(self):
        return self._import_prefix

    @import_prefix.setter
    def import_prefix(self, prefix):
        self._import_prefix = prefix

    @property
    def export_if_missing(self):
        return self._export_if_missing

    @export_if_missing.setter
    def export_if_missing(self, toggle):
        self._export_if_missing = toggle


CONFIG = Config()



"""
.. module: historical_reports.s3.util
    :platform: Unix
    :copyright: (c) 2017 by Netflix Inc., see AUTHORS for more
    :license: Apache, see LICENSE for more details.
.. author:: Mike Grima <mgrima@netflix.com>
"""
import boto3
from botocore.exceptions import ClientError
from retrying import retry

from historical_reports.s3.config import CONFIG

import logging

logging.basicConfig()
log = logging.getLogger('historical-reports-s3')
log.setLevel(logging.INFO)


@retry(stop_max_attempt_number=3, wait_exponential_multiplier=1000, wait_exponential_max=10000)
def _upload_to_s3(file, client, bucket, prefix, content_type="application/json"):
    client.put_object(Bucket=bucket, Key=prefix, Body=file, ContentType=content_type)


@retry(stop_max_attempt_number=3, wait_exponential_multiplier=1000, wait_exponential_max=10000)
def _get_from_s3(client, bucket, prefix):
    try:
        return client.get_object(Bucket=bucket, Key=prefix)["Body"].read().decode()
    except ClientError as ce:
        if ce.response['Error']['Code'] == 'NoSuchKey':
            return None


def dump_to_s3(file):
    """
    This will dump the generated schema to S3.
    :param blob:
    :return:
    """
    client = boto3.client("s3")

    # Loop over each bucket and dump:
    for bucket in CONFIG.dump_to_buckets:
        log.debug("[ ] Dumping to {}/{}".format(bucket, CONFIG.dump_to_prefix))
        _upload_to_s3(file, client, bucket, CONFIG.dump_to_prefix)
        log.debug("[+] Complete")

    log.debug("[+] Completed dumping to all buckets.")


def fetch_from_s3():
    """
    This will fetch the report object from S3.
    :param bucket:
    :param prefix:
    :return:
    """
    client = boto3.client("s3")
    return _get_from_s3(client, CONFIG.import_bucket, CONFIG.import_prefix)


def set_config_from_input(lambda_input):
    """
    Sets the attributes on the configuration based on the input to the lambda function.

    These need to be set to the raw values that the configuration needs. For example, if the env var
    would take in a comma-separated-list, you would supply an actual list of the items, not the comma-separated-string.

    TODO: Make this a decorator.
    :param lambda_input:
    :return:
    """
    if lambda_input.get("config"):
        for attribute, value in lambda_input["config"].items():
            if hasattr(CONFIG, attribute):
                setattr(CONFIG, attribute, value)
            else:
                log.error("Config Attribute: {} is not valid.".format(attribute))



"""
.. module: s3-report.tests.factories
    :platform: Unix
    :copyright: (c) 2017 by Netflix Inc., see AUTHORS for more
    :license: Apache, see LICENSE for more details.
.. author:: Kevin Glisson <kglisson@netflix.com>
"""
# Copy and pasted the factories from Historical.
import datetime

from boto3.dynamodb.types import TypeSerializer
from factory import Factory, SubFactory, post_generation
from factory.fuzzy import FuzzyText

seria = TypeSerializer()


def serialize(obj):
    """JSON serializer for objects not serializable by default json code"""

    if isinstance(obj, datetime.datetime):
        serial = obj.replace(microsecond=0).replace(tzinfo=None).isoformat() + "Z"
        return serial

    if isinstance(obj, bytes):
        return obj.decode('utf-8')

    return obj.__dict__


class Records(object):
    def __init__(self, records):
        self.Records = records


class SessionIssuer(object):
    def __init__(self, userName, type, arn, principalId, accountId):
        self.userName = userName
        self.type = type
        self.arn = arn
        self.principalId = principalId
        self.accountId = accountId


class SessionIssuerFactory(Factory):
    class Meta:
        model = SessionIssuer

    userName = FuzzyText()
    type = 'Role'
    arn = 'arn:aws:iam::123456789012:role/historical_poller'
    principalId = 'AROAIKELBS2RNWG7KASDF'
    accountId = '123456789012'


class UserIdentity(object):
    def __init__(self, sessionContext, principalId, type):
        self.sessionContext = sessionContext
        self.principalId = principalId
        self.type = type


class UserIdentityFactory(Factory):
    class Meta:
        model = UserIdentity

    sessionContext = SubFactory(SessionIssuerFactory)
    principalId = 'AROAIKELBS2RNWG7KASDF:joe@example.com'
    type = 'Service'


class DynamoDBData(object):
    def __init__(self, NewImage, OldImage, Keys):
        self.OldImage = {k: seria.serialize(v) for k, v in OldImage.items()}
        self.NewImage = {k: seria.serialize(v) for k, v in NewImage.items()}
        self.Keys = {k: seria.serialize(v) for k, v in Keys.items()}


class DynamoDBDataFactory(Factory):
    class Meta:
        model = DynamoDBData

    NewImage = {}
    Keys = {}
    OldImage = {}


class DynamoDBRecord(object):
    def __init__(self, dynamodb, eventName, userIdentity):
        self.dynamodb = dynamodb
        self.eventName = eventName
        self.userIdentity = userIdentity


class DynamoDBRecordFactory(Factory):
    """Factory generating a DynamoDBRecord"""
    class Meta:
        model = DynamoDBRecord

    dynamodb = SubFactory(DynamoDBDataFactory)
    eventName = 'INSERT'
    userIdentity = SubFactory(UserIdentityFactory)


class DynamoDBRecordsFactory(Factory):
    class Meta:
        model = Records

    @post_generation
    def Records(self, create, extracted, **kwargs):
        if not create:
            # Simple build, do nothing.
            return

        if extracted:
            # A list of groups were passed in, use them
            for record in extracted:
                self.Records.append(record)



"""
.. module: historical_reports.s3.tests.test_everything
    :platform: Unix
    :copyright: (c) 2017 by Netflix Inc., see AUTHORS for more
    :license: Apache, see LICENSE for more details.
.. author:: Mike Grima <mgrima@netflix.com>
"""
import json

import pytest
from historical_reports.s3.entrypoints import handler
from historical_reports.s3.config import CONFIG
from historical_reports.s3.generate import dump_report
from historical.s3.models import CurrentS3Model
from historical_reports.s3.models import S3ReportSchema
from historical_reports.s3.update import process_dynamodb_record, update_records
from historical_reports.s3.util import dump_to_s3, set_config_from_input


class MockContext:
    def get_remaining_time_in_millis(self):
        return 9000


def test_historical_table_fixture(historical_table):
    assert CurrentS3Model.count() == 10


def test_dump_buckets_fixture(dump_buckets):
    assert len(dump_buckets.list_buckets()["Buckets"]) == 10


def test_generated_file_fixture_is_json(generated_file):
    assert type(generated_file) is bytes
    assert len(json.loads(generated_file.decode("utf-8"))["buckets"]) == 10


def test_existing_s3_report_fixture(existing_s3_report, dump_buckets, generated_file):
    assert generated_file == dump_buckets.get_object(Bucket="dump0", Key="historical-s3-report.json")["Body"].read()


def test_bucket_schema(historical_table):
    all_buckets = CurrentS3Model.scan()
    generated_file = S3ReportSchema(strict=True).dump({"all_buckets": all_buckets}).data

    assert generated_file["generated_date"]
    assert generated_file["s3_report_version"] == CONFIG.s3_reports_version
    assert not generated_file.get("all_buckets")

    for name, value in generated_file["buckets"].items():
        assert value["AccountId"] == "123456789012"
        assert value["Region"] == "us-east-1"
        assert value["Tags"]["theBucketName"] == name
        assert not value.get("_version")
        assert not value.get("Name")


def test_light_bucket_schema(historical_table):
    old_fields = CONFIG.exclude_fields
    CONFIG.exclude_fields = "Name,_version,Grants,LifecycleRules,Logging,Policy,Tags,Versioning,Website,Cors," \
                            "Notifications,Acceleration,Replication,CreationDate,AnalyticsConfigurations," \
                            "MetricsConfigurations,InventoryConfigurations".split(",")

    all_buckets = CurrentS3Model.scan()
    generated_file = S3ReportSchema(strict=True).dump({"all_buckets": all_buckets}).data

    assert generated_file["generated_date"]
    assert generated_file["s3_report_version"] == CONFIG.s3_reports_version
    assert len(generated_file["buckets"]) == 10
    assert not generated_file.get("all_buckets")

    for bucket in generated_file["buckets"].values():
        keys = bucket.keys()
        for excluded in CONFIG.exclude_fields:
            assert excluded not in keys

        assert bucket["AccountId"] == "123456789012"
        assert bucket["Region"] == "us-east-1"

    # Clean-up:
    CONFIG.exclude_fields = old_fields


def test_dump_to_s3(dump_buckets, generated_file):
    old_value = CONFIG.dump_to_buckets
    CONFIG.dump_to_buckets = ["dump{}".format(x) for x in range(0, 10)]

    dump_to_s3(generated_file)

    # Check that it's all good:
    for bucket in CONFIG.dump_to_buckets:
        assert dump_buckets.get_object(Bucket=bucket, Key=CONFIG.dump_to_prefix)["Body"].read() == generated_file

    # Clean-up:
    CONFIG.dump_to_buckets = old_value


@pytest.mark.parametrize("lambda_entry", [False, True])
def test_dump_report(dump_buckets, historical_table, lambda_entry):
    old_value = CONFIG.dump_to_buckets
    CONFIG.dump_to_buckets = ["dump{}".format(x) for x in range(0, 10)]

    if lambda_entry:
        handler({}, MockContext())
    else:
        dump_report()

    # Verify all the info:
    for bucket in CONFIG.dump_to_buckets:
        file = json.loads(dump_buckets.get_object(Bucket=bucket, Key=CONFIG.dump_to_prefix)["Body"].read().decode())

        assert file["generated_date"]
        assert file["s3_report_version"] == CONFIG.s3_reports_version
        assert not file.get("all_buckets")

        for name, value in file["buckets"].items():
            assert value["AccountId"] == "123456789012"
            assert value["Region"] == "us-east-1"
            assert value["Tags"]["theBucketName"] == name
            assert not value.get("_version")
            assert not value.get("Name")

    # Clean-up:
    CONFIG.dump_to_buckets = old_value


@pytest.mark.parametrize("change_type", ["INSERT", "MODIFY"])
def test_process_dynamodb_record(bucket_event, generated_report, change_type):
    bucket_event["Records"][0]["eventName"] = change_type
    generated_report["all_buckets"] = []

    process_dynamodb_record(bucket_event["Records"][0], generated_report)

    assert len(generated_report["all_buckets"]) == 1
    assert generated_report["all_buckets"][0].Region == "us-east-1"


def test_process_dynamodb_record_deletion(delete_bucket_event, generated_report):
    generated_report["all_buckets"] = []
    process_dynamodb_record(delete_bucket_event["Records"][0], generated_report)

    # Should not do anything -- since not present in the list:
    assert not generated_report["all_buckets"]

    # Check if removal logic works:
    generated_report["buckets"]["testbucketNEWBUCKET"] = {"some configuration": "this should be deleted"}

    # If we receive a removal event that is NOT from a TTL, that should not have any effect:
    delete_bucket_event["Records"][0]["eventName"] = "REMOVE"
    process_dynamodb_record(delete_bucket_event["Records"][0], generated_report)
    assert generated_report["buckets"]["testbucketNEWBUCKET"]

    # Standard "MODIFY" for deletion:
    delete_bucket_event["Records"][0]["eventName"] = "MODIFY"
    process_dynamodb_record(delete_bucket_event["Records"][0], generated_report)
    assert not generated_report["buckets"].get("testbucketNEWBUCKET")


def test_process_dynamodb_record_ttl(ttl_event, generated_report):
    generated_report["all_buckets"] = []
    process_dynamodb_record(ttl_event["Records"][0], generated_report)

    # Should not do anything -- since not present in the list:
    assert not generated_report["all_buckets"]

    generated_report["buckets"]["testbucketNEWBUCKET"] = {"some configuration": "this should be deleted"}
    process_dynamodb_record(ttl_event["Records"][0], generated_report)
    assert not generated_report["buckets"].get("testbucketNEWBUCKET")


def test_bucket_schema_for_events(historical_table, generated_report, bucket_event):
    generated_report["all_buckets"] = []
    process_dynamodb_record(bucket_event["Records"][0], generated_report)

    full_report = S3ReportSchema(strict=True).dump(generated_report).data

    assert full_report["generated_date"]
    assert full_report["s3_report_version"] == CONFIG.s3_reports_version
    assert not full_report.get("all_buckets")

    assert full_report["buckets"]["testbucketNEWBUCKET"]
    assert len(full_report["buckets"]) == 11

    for name, value in full_report["buckets"].items():
        assert value["AccountId"] == "123456789012"
        assert value["Region"] == "us-east-1"
        assert value["Tags"]["theBucketName"] == name
        assert not value.get("_version")
        assert not value.get("Name")


def test_lite_bucket_schema_for_events(historical_table, bucket_event):
    old_fields = CONFIG.exclude_fields
    CONFIG.exclude_fields = "Name,Owner,_version,Grants,LifecycleRules,Logging,Policy,Tags,Versioning,Website,Cors," \
                            "Notifications,Acceleration,Replication,CreationDate,AnalyticsConfigurations," \
                            "MetricsConfigurations,InventoryConfigurations".split(",")

    all_buckets = CurrentS3Model.scan()
    generated_report = S3ReportSchema(strict=True).dump({"all_buckets": all_buckets}).data

    generated_report["all_buckets"] = []
    process_dynamodb_record(bucket_event["Records"][0], generated_report)

    lite_report = S3ReportSchema(strict=True).dump(generated_report).data

    assert lite_report["generated_date"]
    assert lite_report["s3_report_version"] == CONFIG.s3_reports_version
    assert not lite_report.get("all_buckets")

    assert lite_report["buckets"]["testbucketNEWBUCKET"]
    assert len(lite_report["buckets"]) == 11

    for bucket in lite_report["buckets"].values():
        keys = bucket.keys()
        for excluded in CONFIG.exclude_fields:
            assert excluded not in keys

        assert bucket["AccountId"] == "123456789012"
        assert bucket["Region"] == "us-east-1"

    # Clean-up:
    CONFIG.exclude_fields = old_fields


@pytest.mark.parametrize("lambda_entry", [False, True])
def test_update_records(existing_s3_report, historical_table, bucket_event, delete_bucket_event, dump_buckets,
                        lambda_entry):
    old_dump_to_buckets = CONFIG.dump_to_buckets
    old_import_bucket = CONFIG.import_bucket

    CONFIG.import_bucket = "dump0"
    CONFIG.dump_to_buckets = ["dump0"]

    # Add a bucket:
    if lambda_entry:
        handler(bucket_event, MockContext())
    else:
        update_records(bucket_event["Records"])

    new_report = json.loads(
        dump_buckets.get_object(Bucket="dump0", Key="historical-s3-report.json")["Body"].read().decode("utf-8")
    )
    assert len(new_report["buckets"]) == 11

    existing_json = json.loads(existing_s3_report.decode("utf-8"))
    assert len(new_report["buckets"]) != len(existing_json["buckets"])
    assert new_report["buckets"]["testbucketNEWBUCKET"]

    # Delete a bucket:
    if lambda_entry:
        handler(delete_bucket_event, MockContext())
    else:
        update_records(delete_bucket_event["Records"])

    delete_report = json.loads(
        dump_buckets.get_object(Bucket="dump0", Key="historical-s3-report.json")["Body"].read().decode("utf-8")
    )
    assert len(delete_report["buckets"]) == len(existing_json["buckets"]) == 10
    assert not delete_report["buckets"].get("testbucketNEWBUCKET")

    # Clean-up:
    CONFIG.dump_to_buckets = old_dump_to_buckets
    CONFIG.import_bucket = old_import_bucket


def test_update_records_sans_existing(historical_table, dump_buckets, bucket_event):
    old_dump_to_buckets = CONFIG.dump_to_buckets
    old_import_bucket = CONFIG.import_bucket
    old_export_if_missing = CONFIG.export_if_missing

    # First test that the object is missing, and we aren't going to perform a full report dump:
    CONFIG.import_bucket = "dump0"
    CONFIG.export_if_missing = False
    update_records(bucket_event["Records"])
    assert not dump_buckets.list_objects_v2(Bucket="dump0")["KeyCount"]

    # Now, with commit set to False:
    CONFIG.export_if_missing = True
    update_records(bucket_event["Records"], commit=False)
    assert not dump_buckets.list_objects_v2(Bucket="dump0")["KeyCount"]

    # Now with commit:
    update_records(bucket_event["Records"])
    assert len(dump_buckets.list_objects_v2(Bucket="dump0")["Contents"]) == 1

    # Clean-up:
    CONFIG.dump_to_buckets = old_dump_to_buckets
    CONFIG.import_bucket = old_import_bucket
    CONFIG.export_if_missing = old_export_if_missing


def test_configuration_update():
    # Just test that a handful are working:
    old_import_prefix = CONFIG.import_prefix
    old_current_region = CONFIG.current_region
    old_exclude_fields = CONFIG.exclude_fields

    lambda_config = {
        "config": {
            "import_prefix": "some-prefix",
            "exclude_fields": ["all", "of", "them"],
            "current_region": "us-west-2"
        }
    }

    set_config_from_input(lambda_config)

    assert CONFIG.import_prefix == "some-prefix"
    assert CONFIG.current_region == "us-west-2"
    assert CONFIG.exclude_fields == ["all", "of", "them"]

    # Clean-up:
    CONFIG.import_prefix = old_import_prefix
    CONFIG.current_region = old_current_region
    CONFIG.exclude_fields = old_exclude_fields



"""
.. module: historical_reports.s3.entrypoints
    :platform: Unix
    :copyright: (c) 2017 by Netflix Inc., see AUTHORS for more
    :license: Apache, see LICENSE for more details.
.. author:: Mike Grima <mgrima@netflix.com>
"""
from raven_python_lambda import RavenLambdaWrapper
from historical_reports.s3.generate import dump_report
from historical_reports.s3.update import update_records
from historical_reports.s3.util import set_config_from_input


@RavenLambdaWrapper()
def handler(event, context):
    """
    Historical S3 report generator lambda handler. This will handle both scheduled events as well as dynamo stream
    events.
    """
    set_config_from_input(event)

    if event.get("Records"):
        # Update event:
        update_records(event["Records"])

    else:
        # Generate event:
        dump_report()



"""
Historical Reports
==================

A collection of reports using data collected by Historical.

"""
import sys
import os.path

from setuptools import setup, find_packages


ROOT = os.path.realpath(os.path.join(os.path.dirname(__file__)))

# When executing the setup.py, we need to be able to import ourselves, this
# means that we need to add the src/ directory to the sys.path.
sys.path.insert(0, ROOT)

about = {}
with open(os.path.join(ROOT, "historical_reports", "__about__.py")) as f:
    exec(f.read(), about)


install_requires = [
    'historical==0.1.6',
    'retrying==1.3.3',
    'click==6.7'
]

tests_require = [
    'pytest==3.1.3',
    'moto>=1.1.24',
    'coveralls==1.1',
    'factory-boy==2.9.2',
    'tox==2.9.1'
]

setup(
    name=about["__title__"],
    version=about["__version__"],
    author=about["__author__"],
    author_email=about["__email__"],
    url=about["__uri__"],
    description=about["__summary__"],
    long_description='See README.md',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    extras_require={
        'tests': tests_require
    },
    entry_points={
        'console_scripts': [
            's3report = historical_reports.s3.cli:cli',
        ]
    },
    keywords=['aws', 'account_management', "s3"]
)



"""
.. module: historical_reports.s3.models
    :platform: Unix
    :copyright: (c) 2017 by Netflix Inc., see AUTHORS for more
    :license: Apache, see LICENSE for more details.
.. author:: Mike Grima <mgrima@netflix.com>
"""
from marshmallow import Schema, fields
from marshmallow.fields import Field
from datetime import datetime

from historical_reports.s3.config import CONFIG

import logging

logging.basicConfig()
log = logging.getLogger('historical-reports-s3')
log.setLevel(logging.INFO)


def get_generated_time(*args):
    return datetime.utcnow().replace(tzinfo=None, microsecond=0).isoformat() + "Z"


def _serialize_bucket(bucket, account_id, region):
    bucket["AccountId"] = account_id
    bucket["Region"] = region

    # Remove fields in the exclusion list:
    for e in CONFIG.exclude_fields:
        bucket.pop(e, None)

    return bucket


class BucketField(Field):
    def _serialize(self, value, attr=None, data=None):
        buckets = data.get("buckets", {})
        for b in data["all_buckets"]:
            log.debug("Fetched details for bucket: {}".format(b.arn))
            name = b.configuration.attribute_values["Name"]  # Store because removed by default.

            # Add the bucket:
            buckets[name] = _serialize_bucket(b.configuration.attribute_values, b.accountId, b.Region)

        return buckets

    def _deserialize(self, value, attr, data):
        return {name: details for name, details in data["buckets"].items()}


class S3ReportSchema(Schema):
    s3_report_version = fields.Int(dump_only=True, required=True, default=CONFIG.s3_reports_version)
    generated_date = fields.Function(get_generated_time, required=True)
    all_buckets = BucketField(required=True, dump_to="buckets")
    buckets = BucketField(required=True, load_from="buckets", load_only=True)



"""
.. module: historical_reports.s3.cli
    :platform: Unix
    :copyright: (c) 2017 by Netflix Inc., see AUTHORS for more
    :license: Apache, see LICENSE for more details.
.. author:: Mike Grima <mgrima@netflix.com>
"""
import logging

import click

from historical_reports.s3.config import CONFIG
from historical_reports.s3.generate import dump_report

logging.basicConfig()
log = logging.getLogger('historical-reports-s3')
log.setLevel(logging.DEBUG)


@click.group()
def cli():
    """Historical-Reports commandline for managing historical reports."""
    pass


def get_exclude_fields(ctx, param, fields):
    CONFIG.exclude_fields = fields.split(",")


def get_bucket(ctx, param, buckets):
    CONFIG.dump_to_buckets = buckets.split(",")


def get_dump_prefix(ctx, param, prefix):
    CONFIG.dump_to_prefix = prefix


@cli.command()
@click.option("--bucket", type=click.STRING, required=True, help="Comma separated list of S3 bucket to dump the "
                                                                 "report to.", callback=get_bucket)
@click.option("--exclude-fields", type=click.STRING, required=False, default="Name,_version",
              help="Comma separated top-level fields to not be included in the final report.",
              callback=get_exclude_fields)
@click.option("--dump-prefix", type=click.STRING, required=False, default="historical-s3-report.json",
              callback=get_dump_prefix)
@click.option("-c", "--commit", default=False, is_flag=True, help="Will only dump to S3 if commit flag is present")
def generate(bucket, exclude_fields, dump_prefix, commit):
    if not commit:
        log.warning("COMMIT FLAG NOT SET -- NOT SAVING ANYTHING TO S3!")
    dump_report(commit=commit)
